/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   mumps.c
 *    main module of freem
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdlib.h>
#include <stddef.h>
#include "mpsdef.h"
#include "errmsg.h"
#include "iniconf.h"
#include "namespace.h"
#include "transact.h"
#include "init.h"
#include <setjmp.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <limits.h>
#include <sys/types.h>
#include <pwd.h>
#include <string.h>

#include "version.h"

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX) && !defined(__osf__) && !defined(MSDOS)
# include <getopt.h>
#endif

#if !defined(PATH_MAX) && defined(_SCO_DS)
# define PATH_MAX 1024
#endif

extern int xecline(int typ);
extern char *getenv(const char *name);
void freem_usage(void);
void freem_print_version(void);

int main (int argc, char **argv, char **envp)
{

    short dx_mode = 0;
    
    int c;
    int import_env = FALSE;

    int option_index = 0;

    char dx_mcode[512];
    
#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX) && !defined(__osf__) && !defined(MSDOS)
    struct option long_options[] = {       
        {"help", no_argument, 0, 'h'},
        {"filter", no_argument, &filter, TRUE},
        {"import", no_argument, &import_env, TRUE},
        {"quiet", no_argument, &quiet_mode, TRUE},
        {"routine", required_argument, 0, 'r'},
        {"namespace", required_argument, 0, 'n'},
        {"config", required_argument, 0, 'c'},
	{"version", no_argument, 0, 'v'},
	{"execute", required_argument, 0, 'x'},
        {0, 0, 0, 0}
    };
#endif
    char nsnbuf[256];
    
    char **env;      
    char *varname = (char *) calloc(STRLEN, sizeof(char));
    char *varval = (char *) calloc(STRLEN, sizeof(char));

    char *symname = (char *) calloc(STRLEN, sizeof(char));
    char *symval = (char *) calloc(STRLEN, sizeof(char));

    int namelen;
    int vallen;

    char cli_rtn_path[PATH_MAX];
    char *cli_rtn_file;
    char cli_rtn_name[256];

    m_argc = argc;			/* save arguments count     */
    m_argv = argv;			/* save arguments string    */
    m_envp = envp;			/* save environment pointer */

    strcpy (zb, argv[0]);		/* name with which mumps has been called */
    stcnv_c2m (zb);
    strcpy (stack0, argv[0]);
    stcnv_c2m (stack0);

    strncpy (config_file, SYSCONFDIR"/freem.conf", 100);
 
#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX) && !defined(__osf__) && !defined(MSDOS)

    while(1) {
        
        c = getopt_long(argc, argv, "hfiqr:n:c:vx:", long_options, &option_index);

        if (c == -1) break;
	if (c == '?') freem_usage ();
	
        switch(c) {
            case 'h':
                freem_usage ();
                break;

            case 'f':
                filter = TRUE;
                break;

            case 'i':
                import_env = TRUE;
                break;

            case 'q':
                quiet_mode = TRUE;
                break;

            case 'c': /* specify FreeM configuration file */
                strncpy (config_file, optarg, 4095);
                break;

            case 'r': /* startup routine */
                direct_mode = FALSE;
                usermode = 0;

                strcpy (startuprou, optarg);
                startuprou[strlen (startuprou)] = '\201';
                break;

            case 'n': /* namespace */
                strcpy (nsname, optarg);               
                break;

	    case 'v': /* version */
		freem_print_version ();
		break;

	    case 'x': /* execute */
		direct_mode = FALSE;
		usermode = 0;
		dx_mode = 1;
		
		strncpy (startuprou, "^%DX\201", 10);
		strncpy (dx_mcode, optarg, 512 - 1);

		stcnv_c2m (dx_mcode);

		break;
		
        } 

    }

#else
    
    {
        extern char *optarg;
        extern int optind, optopt;

        while ((c = getopt(argc, argv, "hfCsiqr:n:c:vx:")) != -1) {

	    if (c == '?') freem_usage ();
	    
            switch(c) {

		case 'h':
		    freem_usage ();
		    break;
		
                case 'f':
                    filter = TRUE;
                    break;

                case 'i':
                    import_env = TRUE;
                    break;

                case 'q':
                    quiet_mode = TRUE;
                    break;

                case 'c':  /* specify FreeM configuration file */
                    strncpy (config_file, optarg, 4095);
                    break;

                case 'r': /* startup routine */
                    direct_mode = FALSE;
                    usermode = 0;

                    strcpy (startuprou, optarg);
                    startuprou[strlen (startuprou)] = '\201';
                    break;

                case 'n': /* namespace */
                    strcpy (nsname, optarg);               
                    break;

		case 'v':
		    freem_print_version ();
		    break;

		case 'x': /* execute */
		    direct_mode = FALSE;
		    usermode = 0;
		    dx_mode = 1;
		    
		    strncpy (startuprou, "^%DX\201", 10);
		    strncpy (dx_mcode, optarg, 512 - 1);
		    
		    stcnv_c2m (dx_mcode);

		break;


            } 
        }
    }
#endif

#if !defined(_AIX)
    if(import_env == TRUE) {

        for(env = envp; *env != 0; env++) {
            
            namelen = 0;
            vallen = 0;
            
            varname = strtok(*env, "=");
            varval = strtok(NULL, "=");

            if(varval != NULL) {
                namelen = strlen(varname);
                vallen = strlen(varval);

                strncpy(symname, varname, STRLEN);
                strncpy(symval, varval, STRLEN);

                stcnv_m2c (symname);
                stcnv_m2c (symval);

                symtab(set_sym, symname, symval);
            }
        }
    }
#endif    

    /* handle passing of an arbitrary .m file on the command line */
    /* this is most often used for shebang-line scripts. */
    if (optind < argc) {

        /* bail if file does not exist */
        if (access (argv[optind], F_OK) == -1) {

            set_io (UNIX);
            
            fprintf (stderr, "Routine %s does not exist.\n", argv[optind]);

            exit (1);

        }

        direct_mode = FALSE;
        usermode = 0;

 

        /* was a path specified at all? */
        if (strchr (argv[optind], '/') == NULL) {

            /* the entirety of argv[optind] is the filename */
            cli_rtn_file = argv[optind];            

            /* use the current directory */
            sprintf (cli_rtn_path, ".");

        }
        else {

            /* isolate the filename from the path */
            cli_rtn_file = strrchr (argv[optind], '/') + 1;
            
            /* isolate the routine name from the filename */
            strncpy (cli_rtn_name, cli_rtn_file, strchr (cli_rtn_file, '.') - cli_rtn_file);
            
            /* isolate the path from the routine file */
            strncpy (cli_rtn_path, argv[optind], strrchr (argv[optind], '/') - argv[optind]);

        }

        /* do we have a file extension? */
        if (strchr (cli_rtn_file, '.') != NULL) {
            
            /* if so, just remove it */
            strncpy (cli_rtn_name, cli_rtn_file, strchr (cli_rtn_file, '.') - cli_rtn_file);

        }
        else {

            /* otherwise, just take a direct copy */
            strcpy (cli_rtn_name, cli_rtn_file);

        }

        /* make this the startup routine */
        snprintf (startuprou, 256, "^%s\201", cli_rtn_name);

        /* re-work the namespace config to search for the 
           routine in the discovered path */
        if (cli_rtn_name[0] == '%') {

            snprintf (rou0plib, 256, "%s\201", cli_rtn_path);
            snprintf (rou1plib, 256, "%s\201", cli_rtn_path);

        }
        else {

            snprintf (rou0path, 256, "%s\201", cli_rtn_path);
            snprintf (rou1path, 256, "%s\201", cli_rtn_path);

        }

    }

    strncpy (nsnbuf, nsname, 255);    

#ifdef __EMSCRIPTEN__
    fprintf (stderr, "EMSCRIPTEN CHECK PRE-INIT\n");
#endif
    
    if (!file_exists (config_file)) {

        set_io (UNIX);
        fprintf (stderr, "\nFreeM has not been configured. Please run 'fmadm configure'.\n\n\n\n");
        
        exit (2);

    }
    
    /* initialize FreeM environment */
    if (init (nsnbuf) == FALSE) {

        set_io (UNIX);
        fprintf (stderr, "\nError initializing FreeM.\n");

        exit (1);

    }

    if (direct_mode == TRUE && quiet_mode == FALSE) {

	char verstr[500];
        char version[256];
	
	stcpy (verstr, FREEM_VERSION_STR);
	stcnv_m2c (verstr);
	        
        snprintf (version, 256, "Coherent Logic Development FreeM version %s\r\n\201", verstr);
	write_m (version);

	snprintf (version, 256, "Copyright (C) 2014, 2020, 2021 Coherent Logic Development LLC\r\n\r\n\201");
	write_m (version);

    }
    else {
        write_m ("\r\n\r\n\201");
    }

    if (dx_mode) {
	char k_buf[512];	
	char t_buf[512];
	
	snprintf (k_buf, 512 - 1, "%%DX\201\201");
	symtab (set_sym, k_buf, dx_mcode);	
    }

#ifndef __EMSCRIPTEN__
    /* run mumps */
    xecline (1);
#else
    fprintf (stderr, "EMSCRIPTEN CHECK PRE-XECLINE\n");
    stcpy (code, "QUIT\201");
    xecline (4);
#endif

    exit (0);	/* we should never reach that statement */

} /* end of main() */

void freem_usage(void)
{
    fprintf (stdout, "\nusage:  freem [OPTION...]\n\n");

    fprintf (stdout, "OPTIONS:\n\n");
    fprintf (stdout, "\t-c, --config\n\t\tspecify a configuration file other than %s\n\n", config_file);
    fprintf (stdout, "\t-h, --help\n\t\tdisplays this help message\n\n");
    fprintf (stdout, "\t-i, --import\n\t\timports UNIX environment variables as M locals\n\n");
    fprintf (stdout, "\t-f, --filter\n\t\tallows M code to be used as a filter\n\n");
    fprintf (stdout, "\t-n <NAMESPACE>, --namespace=<NAMESPACE>\n\t\tselects <NAMESPACE> as the startup namespace instead of USER\n\n");
    fprintf (stdout, "\t-q, --quiet\n\t\tdisables startup messages and prompt string\n\n");
    fprintf (stdout, "\t-r <LABEL^ROUTINE>, --routine=<LABEL^ROUTINE>\n\t\texecute <LABEL^ROUTINE> on startup instead of entering direct mode\n\n");
    fprintf (stdout, "\t-v, --version\n\t\tdisplay FreeM version information\n\n");
    fprintf (stdout, "\t-x <MCODE>, --execute=<MCODE>\n\t\texecute M code <MCODE> on startup\n\n");
    fprintf (stdout, "Report bugs to: freem-bugs@coherent-logic.com\n");
    fprintf (stdout, "FreeM home page: <https://freem.coherent-logic.com>\n\n");   
    
    exit (1);
}

void freem_print_version(void)
{
    char verstr[500];
    stcpy (verstr, FREEM_VERSION_STR);
    stcnv_m2c (verstr);
    
    fprintf (stdout, "Coherent Logic Development FreeM %s\n", verstr);
    fprintf (stdout, "Copyright (C) 2014, 2020, 2021 Coherent Logic Development LLC\n");
    fprintf (stdout, "License AGPLv3+: GNU AGPL version 3 or later <https://gnu.org/license/agpl-3.0.html>\n");
    fprintf (stdout, "This is free software: you are free to change and redistribute it.\n");
    fprintf (stdout, "There is NO WARRANTY, to the extent permitted by law.\n");
    
    exit (0);
}
